/*
   (c) Copyright 2001-2007  The DirectFB Organization (directfb.org)
   (c) Copyright 2000-2004  Convergence (integrated media) GmbH

   All rights reserved.

   Written by Antonino Daplas <adaplas@pol.net>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <config.h>

#include <fbdev/fbdev.h>  /* FIXME: Needs to be included before dfb_types.h to work around a type clash with asm/types.h */
#include <dfb_types.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <sys/mman.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <malloc.h>

#include <directfb.h>

#include <core/coredefs.h>
#include <core/coretypes.h>
#include <core/screens.h>

#include <core/state.h>
#include <core/gfxcard.h>
#include <core/surface_buffer.h>

#include <gfx/convert.h>
#include <gfx/util.h>
#include <misc/conf.h>
#include <misc/util.h>

#include <core/graphics_driver.h>

#include <sys/time.h>

DFB_GRAPHICS_DRIVER( tcc893x_vioc )

#include "tcc893x_vioc.h"

#if 0
#define dprintf printf
#else
#define dprintf
#endif

#if 0
static inline long mymicroclock()
{
     struct timeval tv;

     gettimeofday (&tv, NULL);

     return tv.tv_sec * 1000000 + tv.tv_usec;
}
#endif

D_DEBUG_DOMAIN( tcc893x_vioc, "tcc893x_vioc", "Telechip TCC893X VIOC" );

#define TCC893X_VIOC_SUPPORTED_DRAWINGFLAGS (DSDRAW_NOFX)

#define TCC893X_VIOC_SUPPORTED_DRAWINGFUNCTIONS (DFXL_NONE)

#define TCC893X_VIOC_SUPPORTED_BLITTINGFLAGS (DSBLIT_SRC_COLORKEY|DSBLIT_BLEND_ALPHACHANNEL)

#define TCC893X_VIOC_SUPPORTED_BLITTINGFUNCTIONS  (DFXL_STRETCHBLIT|DFXL_BLIT)

static inline void tcc893x_vioc_init_local_variables(VIOCDriverData *tcc893x_viocdrv, VIOCDeviceData *tcc893x_viocdev)
{
	//init function status
	tcc893x_viocdev->tcc_dfb_dev.bAlpha = false;
	tcc893x_viocdev->tcc_dfb_dev.bColorkey = false;

	//init color value
	tcc893x_viocdev->tcc_dfb_dev.a = 0x0;
	tcc893x_viocdev->tcc_dfb_dev.r = 0x0;
	tcc893x_viocdev->tcc_dfb_dev.g = 0x0;
	tcc893x_viocdev->tcc_dfb_dev.b = 0x0;
}

static inline void  tcc893x_vioc_set_destination( VIOCDriverData *tcc893x_viocdrv, VIOCDeviceData *tcc893x_viocdev, CardState *state )
{
	CoreSurfaceBuffer *buffer = state->dst.buffer;
	dprintf("%s\r\n", __func__);

	// RGB Only..!
	memset(&tcc893x_viocdev->dst, 0, sizeof(tcc893x_viocdev->dst));

	tcc893x_viocdev->dst.y_addr = state->dst.phys;
	tcc893x_viocdev->dst.pitch = state->dst.pitch;

	tcc893x_viocdev->tcc_dfb_dev.dest_Yaddr = state->dst.phys;
	tcc893x_viocdev->tcc_dfb_dev.dest_offset = state->dst.pitch;
	
	switch (buffer->format) {
	  case DSPF_RGB16:
	  	tcc893x_viocdev->dst.bpp = 2;
	       tcc893x_viocdev->dst.format = DSPF_RGB16;
		tcc893x_viocdev->tcc_dfb_dev.dest_fmt = TCC_LCDC_IMG_FMT_RGB565;
	       break;
	 case DSPF_RGB24:
		tcc893x_viocdev->dst.bpp = 3;
	       tcc893x_viocdev->dst.format = DSPF_RGB24;
		tcc893x_viocdev->tcc_dfb_dev.dest_fmt = TCC_LCDC_IMG_FMT_RGB888_3;
	       break;
	 case DSPF_RGB32:
	 case DSPF_ARGB:
	       tcc893x_viocdev->dst.bpp = 4;
		tcc893x_viocdev->dst.format = DSPF_ARGB;
		tcc893x_viocdev->tcc_dfb_dev.dest_fmt = TCC_LCDC_IMG_FMT_RGB888;
	       break;
	  default:
	       D_BUG("unexpected pixelformat~");
	}
}

static inline void tcc893x_vioc_set_src(VIOCDriverData *tcc893x_viocdrv, VIOCDeviceData *tcc893x_viocdev, CardState      *state)
{
	CoreSurfaceBuffer *buffer = state->src.buffer;
	dprintf("%s\r\n", __func__);

	tcc893x_viocdev->src.y_addr = state->src.phys;
	tcc893x_viocdev->src.pitch = state->src.pitch;

	tcc893x_viocdev->tcc_dfb_dev.src_Yaddr = state->src.phys;
	tcc893x_viocdev->tcc_dfb_dev.src_offset = state->src.pitch;

	switch (buffer->format) {
	case DSPF_RGB16:
		//local
		tcc893x_viocdev->src.bpp = 2;
	       tcc893x_viocdev->src.format = DSPF_RGB16;

		//parameter
		tcc893x_viocdev->tcc_dfb_dev.src_fmt = TCC_LCDC_IMG_FMT_RGB565;
	       break;
	case DSPF_RGB24:
		//local
		tcc893x_viocdev->src.bpp = 3;
	       tcc893x_viocdev->src.format = DSPF_RGB24;

		//parameter  
		tcc893x_viocdev->tcc_dfb_dev.src_fmt = TCC_LCDC_IMG_FMT_RGB888_3;
	       break;
	case DSPF_RGB32:
	case DSPF_ARGB:
		//local
		tcc893x_viocdev->src.bpp = 4;
	       tcc893x_viocdev->src.format = DSPF_ARGB;

		//parameter
		tcc893x_viocdev->tcc_dfb_dev.src_fmt = TCC_LCDC_IMG_FMT_RGB888;
		break;   	
	default:
	       D_BUG("unexpected pixelformat~");
	}

	tcc893x_viocdev->srcpitch = state->src.pitch;
}

static inline void tcc893x_vioc_set_colorkey(VIOCDriverData *tcc893x_viocdrv, VIOCDeviceData *tcc893x_viocdev, CardState *state)
{
	DFBColor color;
	dprintf("%s color(0x%x) fmt(0x%x)\r\n",__func__, state->src_colorkey, tcc893x_viocdev->src.format);

	dfb_pixel_to_color(tcc893x_viocdev->src.format, state->src_colorkey, &color);

	tcc893x_viocdev->tcc_dfb_dev.r = color.r;
	tcc893x_viocdev->tcc_dfb_dev.g = color.g;
	tcc893x_viocdev->tcc_dfb_dev.b = color.b;
	tcc893x_viocdev->tcc_dfb_dev.a = color.a;
}

static void tcc893x_viocCheckState(void *drv, void *dev, CardState *state, DFBAccelerationMask accel )
{
	VIOCDriverData *tcc893x_viocdrv = (VIOCDriverData *) drv;
	VIOCDeviceData *tcc893x_viocdev = (VIOCDeviceData *) dev;
	dprintf("%s\r\n", __func__);
	
	switch (state->destination->config.format) {
	case DSPF_ARGB1555:
	case DSPF_RGB16:
	case DSPF_RGB24:
	case DSPF_RGB32:
	case DSPF_ARGB:
		break;
	default:
		return;
	}

	if (!(accel & ~TCC893X_VIOC_SUPPORTED_DRAWINGFUNCTIONS))
	{
		if(!(state->drawingflags & ~TCC893X_VIOC_SUPPORTED_DRAWINGFLAGS))
		{
	  		state->accel |= TCC893X_VIOC_SUPPORTED_DRAWINGFUNCTIONS;
		}
	}

	if (!(accel & ~TCC893X_VIOC_SUPPORTED_BLITTINGFUNCTIONS))
	{
		if(!(state->blittingflags & ~TCC893X_VIOC_SUPPORTED_BLITTINGFLAGS)) 
		{
			//if (state->source->config.format == state->destination->config.format)
	       		state->accel |= TCC893X_VIOC_SUPPORTED_BLITTINGFUNCTIONS;
			//else
				//dprintf("tcc893x_vioc sfmt(0x%x), dfmt(0x%x)\r\n", state->source->config.format, state->destination->config.format);
		}
	}

}

static void
tcc893x_viocSetState( void *drv, void *dev,
              GraphicsDeviceFuncs *funcs,
              CardState *state, DFBAccelerationMask accel )
{
	VIOCDriverData *tcc893x_viocdrv = (VIOCDriverData *) drv;
	VIOCDeviceData *tcc893x_viocdev = (VIOCDeviceData *) dev;

	StateModificationFlags  modified = state->mod_hw;	
	
	dprintf("%s\r\n", __func__);

	if(!state->src.phys || !state->dst.phys)
	{
		printf("%s failed.. param\r\n", __func__);
		return;
	}
	
	if (modified & SMF_SOURCE && state->source )
	  tcc893x_viocdev->smf_source = 0;

	if (modified & SMF_DESTINATION)
	  tcc893x_viocdev->smf_destination = tcc893x_viocdev->smf_color = 0;

	if (modified & SMF_COLOR)
	  tcc893x_viocdev->smf_color = 0;

	if (modified & SMF_CLIP)
	  tcc893x_viocdev->smf_clip = 0;
	
	tcc893x_vioc_set_destination(tcc893x_viocdrv, tcc893x_viocdev, state);
	tcc893x_vioc_set_src(tcc893x_viocdrv, tcc893x_viocdev, state);
	
	switch (accel) 	{
	case DFXL_STRETCHBLIT:
		state->set = DFXL_STRETCHBLIT;
		tcc893x_viocdev->tcc_dfb_dev.blit_status = DFXL_STRETCHBLIT;
		break;
	case DFXL_BLIT:
		state->set = DFXL_BLIT;
		tcc893x_viocdev->tcc_dfb_dev.blit_status = DFXL_BLIT;
		break;
	default:
		D_ERROR( "unexpected drawing/blitting function" );
		break;
	}

	tcc893x_vioc_init_local_variables(tcc893x_viocdrv, tcc893x_viocdev);

	if(state->blittingflags & DSBLIT_BLEND_ALPHACHANNEL){
		tcc893x_viocdev->tcc_dfb_dev.bAlpha = true;
	}
	
	if(state->blittingflags & DSBLIT_SRC_COLORKEY){
		tcc893x_vioc_set_colorkey(tcc893x_viocdrv,  tcc893x_viocdev, state);
		tcc893x_viocdev->tcc_dfb_dev.bColorkey = true;
	}

	#if 0
	if (state->modified & SMF_CLIP)
	  tcc893x_vioc_set_clip( tcc893x_viocdrv, tcc893x_viocdev, &state->clip);
	#endif

     state->mod_hw = 0;
}

static DFBResult
tcc893x_viocEngineSync(void *drv, void *dev)
{
	dprintf("%s\r\n", __func__);
	return DFB_OK;
}

static void
tcc893x_viocEngineReset(void *drv, void *dev)
{
	dprintf("%s\r\n", __func__);
	//memset((void*)dfb_system_video_memory_virtual(0),0,dfb_gfxcard_memory_length());
}

// raw vioc component setting define
#define VIOC_PHY_ADDR 0x72000000
#define VIOC_PHY_LEN  0x00010400

#define VIOC_RDMA_OFFSET 	(0x1500)		// RDMA17
#define VIOC_WDMA_OFFSET	(0x2F00)		// WDMA5
#define VIOC_WMIX_OFFSET	(0x1E00)		// WMIX6
#define VIOC_SCALER_OFFSET 	(0x2200)		// SCALER2

#define READ_REGISTER(R) (*(volatile unsigned int  * const)(R))
#define WRITE_REGISTER(W, V) *(volatile unsigned int * const)(W) = (V)

#define READ_VIOC_RDMA(R) READ_REGISTER((R)+VIOC_RDMA_OFFSET)
#define READ_VIOC_WDMA(R) READ_REGISTER((R)+VIOC_WDMA_OFFSET)
#define READ_VIOC_WMIX(R) READ_REGISTER((R)+VIOC_WMIX_OFFSET)
#define READ_VIOC_SCALER(R) READ_REGISTER((R)+VIOC_SCALER_OFFSET)

#define WRITE_VIOC_RDMA(W,V) WRITE_REGISTER((W)+VIOC_RDMA_OFFSET, (V))
#define WRITE_VIOC_WDMA(W,V) WRITE_REGISTER((W)+VIOC_WDMA_OFFSET, (V))
#define WRITE_VIOC_WMIX(W,V) WRITE_REGISTER((W)+VIOC_WMIX_OFFSET, (V))
#define WRITE_VIOC_SCALER(W,V) WRITE_REGISTER((W)+VIOC_SCALER_OFFSET, (V))

static bool tcc893x_viocBlit( void *drv, void *dev, DFBRectangle *rect, int dx, int dy ){
	VIOCDriverData *tcc893x_viocdrv = (VIOCDriverData *)drv;
	VIOCDeviceData *tcc893x_viocdev = (VIOCDeviceData *)dev;
	unsigned int uiWidth = (tcc893x_viocdev->tcc_dfb_dev.src_offset/tcc893x_viocdev->src.bpp);

	dprintf("[%s] r(%d %d %d %d) dx: %d dy: %d\r\n", __func__,rect->x, rect->y, rect->w, rect->h, dx, dy);

	tcc893x_viocdev->tcc_dfb_dev.src_winLeft = rect->x;
	tcc893x_viocdev->tcc_dfb_dev.src_winTop = rect->y;
	tcc893x_viocdev->tcc_dfb_dev.src_winRight = rect->x + rect->w;
	tcc893x_viocdev->tcc_dfb_dev.src_winBottom = rect->y + rect->h;
		
	tcc893x_viocdev->tcc_dfb_dev.src_ImgWidth= rect->w;
	tcc893x_viocdev->tcc_dfb_dev.src_ImgHeight= rect->h;
		
	tcc893x_viocdev->tcc_dfb_dev.dest_winLeft = dx;
	tcc893x_viocdev->tcc_dfb_dev.dest_winTop = dy;
	tcc893x_viocdev->tcc_dfb_dev.dest_winRight =dx + rect->w;
	tcc893x_viocdev->tcc_dfb_dev.dest_winBottom = dy + rect->h;
		
	tcc893x_viocdev->tcc_dfb_dev.dest_ImgWidth= rect->w;
	tcc893x_viocdev->tcc_dfb_dev.dest_ImgHeight= rect->h;

	tcc893x_viocdev->tcc_dfb_dev.viqe_onthefly= 0;
	tcc893x_viocdev->tcc_dfb_dev.responsetype = SCALER_POLLING;
		
	ioctl(tcc893x_viocdev->dfb_fd, TCC_DFB_VIOC_IOCTL, &tcc893x_viocdev->tcc_dfb_dev);

	return true;
}

static bool tcc893x_viocStretchBlit( void *driver_data, void *device_data, DFBRectangle *srect, DFBRectangle *drect )
{
	VIOCDriverData *tcc893x_viocdrv = (VIOCDriverData *) driver_data;
	VIOCDeviceData *tcc893x_viocdev = (VIOCDeviceData *) device_data;
	
	dprintf("[%s] s(%d %d %d %d) d(%d %d %d %d)\r\n", __func__, srect->x, srect->y, srect->w, srect->h, drect->x, drect->y, drect->w, drect->h);

	tcc893x_viocdev->tcc_dfb_dev.src_winLeft = srect->x;
	tcc893x_viocdev->tcc_dfb_dev.src_winTop = srect->y;
	tcc893x_viocdev->tcc_dfb_dev.src_winRight = srect->x + srect->w;
	tcc893x_viocdev->tcc_dfb_dev.src_winBottom = srect->y + srect->h;
	
	tcc893x_viocdev->tcc_dfb_dev.src_ImgWidth= srect->w;
	tcc893x_viocdev->tcc_dfb_dev.src_ImgHeight= srect->h;
	
	tcc893x_viocdev->tcc_dfb_dev.dest_winLeft = drect->x;
	tcc893x_viocdev->tcc_dfb_dev.dest_winTop = drect->y;
	tcc893x_viocdev->tcc_dfb_dev.dest_winRight = drect->x + drect->w;
	tcc893x_viocdev->tcc_dfb_dev.dest_winBottom = drect->y + drect->h;
	
	tcc893x_viocdev->tcc_dfb_dev.dest_ImgWidth= drect->w;
	tcc893x_viocdev->tcc_dfb_dev.dest_ImgHeight= drect->h;
	
	tcc893x_viocdev->tcc_dfb_dev.viqe_onthefly= 0;
	tcc893x_viocdev->tcc_dfb_dev.responsetype = SCALER_POLLING;

	ioctl(tcc893x_viocdev->dfb_fd, TCC_DFB_VIOC_IOCTL, &tcc893x_viocdev->tcc_dfb_dev);	
	return true;
}

static int
driver_probe( CoreGraphicsDevice *device )
{
	dprintf("%s\r\n", __func__);
	switch (dfb_gfxcard_get_accelerator( device )) {
	  case FB_ACCEL_TCC893x_VIOC:          /* Telechips TCC893X VIOC */
	       return 1;
	}
	return 0;
}

static void
driver_get_info( CoreGraphicsDevice *device,
                 GraphicsDriverInfo *info )
{
	dprintf("%s\r\n", __func__);

	/* fill driver info structure */
	snprintf( info->name,
	       DFB_GRAPHICS_DRIVER_INFO_NAME_LENGTH,
	       "tcc893x_vioc Driver" );

	snprintf( info->vendor,
	       DFB_GRAPHICS_DRIVER_INFO_VENDOR_LENGTH,
	       "Telechips" );

	snprintf( info->url,
	       DFB_GRAPHICS_DRIVER_INFO_URL_LENGTH,
	       "http://www.telechips.com" );

	snprintf( info->license,
	       DFB_GRAPHICS_DRIVER_INFO_LICENSE_LENGTH,
	       "LGPL" );

	info->version.major = 0;
	info->version.minor = 1;

	info->driver_data_size = sizeof (VIOCDriverData);
	info->device_data_size = sizeof (VIOCDeviceData);
}

static DFBResult
driver_init_driver( CoreGraphicsDevice  *device,
                    GraphicsDeviceFuncs *funcs,
                    void                *driver_data,
                    void                *device_data,
                    CoreDFB             *core )
{
	VIOCDriverData *tcc893x_viocdrv = (VIOCDriverData*) driver_data;    
	VIOCDeviceData *tcc893x_viocdev = (VIOCDeviceData*) device_data;
	unsigned int uiPlugin = 1;

	dprintf("%s\r\n", __func__);
	tcc893x_viocdrv->dfb_fbdev = dfb_system_data();

	if((tcc893x_viocdev->dfb_fd= open("/dev/directfb", O_RDWR)) == -1)
	{
		printf("fail directfb driver open\r\n");
	}

	ioctl(tcc893x_viocdev->dfb_fd, TCC_DFB_PATH_CONNECT, &uiPlugin);

	funcs->CheckState    = tcc893x_viocCheckState;
	funcs->SetState      = tcc893x_viocSetState;
	funcs->EngineSync    = tcc893x_viocEngineSync;
	funcs->EngineReset   = tcc893x_viocEngineReset;
	funcs->FlushTextureCache  = NULL;	//tcc893x_viocFlushTextureCache;

	funcs->FillRectangle = NULL;	//tcc893x_viocFillRectangle;
	funcs->DrawLine      = NULL;	//tcc893x_viocDrawLine;
	funcs->Blit          = tcc893x_viocBlit;	//tcc893x_viocBlit;
	funcs->StretchBlit	 = tcc893x_viocStretchBlit;
	return DFB_OK;
}


static DFBResult
driver_init_device( CoreGraphicsDevice *device,
                    GraphicsDeviceInfo *device_info,
                    void               *driver_data,
                    void               *device_data )
{

	VIOCDeviceData *tcc893x_viocdev = (VIOCDeviceData*) device_data;
	dprintf("%s\r\n", __func__);

	/* fill device info */
	snprintf( device_info->name,
	       DFB_GRAPHICS_DEVICE_INFO_NAME_LENGTH, "tcc893x_vioc" );

	snprintf( device_info->vendor,
	       DFB_GRAPHICS_DEVICE_INFO_VENDOR_LENGTH, "Telechips" );

	device_info->caps.flags    = 0;
	device_info->caps.accel    = TCC893X_VIOC_SUPPORTED_DRAWINGFUNCTIONS |
	                          TCC893X_VIOC_SUPPORTED_BLITTINGFUNCTIONS;

	device_info->caps.drawing  = TCC893X_VIOC_SUPPORTED_DRAWINGFLAGS;
	device_info->caps.blitting = TCC893X_VIOC_SUPPORTED_BLITTINGFLAGS;

	device_info->limits.surface_byteoffset_alignment = 32 * 4;
	device_info->limits.surface_pixelpitch_alignment = 32;

	tcc893x_viocdev->fb_store = false;

	return DFB_OK;

}


static void
driver_close_device( CoreGraphicsDevice *device,
                     void               *driver_data,
                     void               *device_data )
{
	dprintf("%s\r\n", __func__);
}

static void
driver_close_driver( CoreGraphicsDevice *device,
                     void               *driver_data )
{
	unsigned int uiConnect = 0;
	VIOCDriverData *tcc893x_viocdrv = (VIOCDriverData*) driver_data;
	VIOCDeviceData *tcc893x_viocdev = (VIOCDeviceData*) device;
	
	dprintf("%s\r\n", __func__);
	ioctl(tcc893x_viocdev->dfb_fd, TCC_DFB_PATH_CONNECT, &uiConnect);
	tcc893x_viocdrv->dfb_fbdev = NULL;
}
